// WindSimulator.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"

#pragma comment(linker,"\"/manifestdependency:type='win32' \
name='Microsoft.Windows.Common-Controls' version='6.0.0.0' \
processorArchitecture='*' publicKeyToken='6595b64144ccf1df' language='*'\"")

#include <windows.h>
#include <iostream>
#include <thread>
#include <string>
#include <Commctrl.h>
#include <shlobj.h>
#include <deque>

#include <gdiplus.h>
using namespace Gdiplus;
#pragma comment (lib,"Gdiplus.lib")

#include "SharedFileOut.h" // Assetto Corsa
#include "SharedMemory.h" // Project Cars 2
#include "r3e.h" // RaceRoom Racing Experience
#include "SerialClass.h"
#include "resource.h"

#pragma comment( lib, "comctl32.lib" )

using namespace std;

struct _config {
	int com_port = 0;
	int fan_off = 26;
	int fan_min = 0;
	int fan_max = 79;
	int car_spd = 0;
} config;

enum { AC_PHYSICS, AC_GRAPHICS, AC_STATIC, PCARS2, R3E, SHARED_MAX };
#define OVERALLOCATE 100000
struct {
	LPCTSTR name;
	int size; 
	void *pointer;
} shares[SHARED_MAX] = {
	{ "acpmf_physics", sizeof(SPageFilePhysics) },
	{ "acpmf_graphics", sizeof(SPageFileGraphic) },
	{ "acpmf_static", sizeof(SPageFileStatic) },
	{ "$pcars2$", sizeof(SharedMemory) },
	{ "$R3E", sizeof(r3e_shared) }
};

Serial* SP = NULL;
NOTIFYICONDATA tray;
HWND console;
string pathConfig;

ULONG_PTR gdiplusToken;
deque<char> history;
#define X 10
#define Y 170
#define W 464
#define H 82

void log(string message)
{
	message += "\r\n";
	int idx = GetWindowTextLength(console);
	SendMessage(console, EM_SETSEL, idx, idx);
	SendMessage(console, EM_REPLACESEL, 0, (LPARAM)message.c_str());
}

void error(string message)
{
	MessageBox(GetActiveWindow(), message.c_str(), "Error", MB_OK);
	exit(1);
}

char fan(float currentSpeed, float maxSpeed)
{
	float fanSpeed = (float)config.fan_min;
	if (maxSpeed > 1)
	{
		int delta = config.fan_max - config.fan_min;
		fanSpeed += currentSpeed / maxSpeed * delta;
	}
	if (fanSpeed > 79)
		return 79;
	if (fanSpeed < 0)
		return 0;
	return (char)round(fanSpeed);
}

LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message)
	{
		// minimize to systray
		case WM_SYSCOMMAND:
			if (wParam == SC_MINIMIZE)
			{
				ShowWindow(hWnd, SW_HIDE);
				break;
			}
			return DefWindowProc(hWnd, message, wParam, lParam);

		// change COM port
		case WM_COMMAND:
			if (HIWORD(wParam) == CBN_SELCHANGE)
			{
				config.com_port = (int) SendMessage((HWND)lParam, CB_GETCURSEL, 0, 0);
				delete SP;
				SP = NULL;
			}
			break;
		
		// click on systray icon
		case WM_APP + 1:
			if (lParam == WM_LBUTTONDOWN)
			{
				ShowWindow(hWnd, IsWindowVisible(hWnd) ? SW_HIDE : SW_RESTORE);
			}
			break;

		// redraw chart
		case WM_PAINT:
			{
				HDC hdcMem;
				PAINTSTRUCT ps;
				BeginPaint(hWnd, &ps);
				RECT rc;
				GetClientRect(hWnd, &rc);
				hdcMem = CreateCompatibleDC(ps.hdc);
				HBITMAP hbmMem = CreateCompatibleBitmap(ps.hdc,
					rc.right - rc.left,
					rc.bottom - rc.top);
				HBITMAP hbmOld = (HBITMAP) SelectObject(hdcMem, hbmMem);

				HBRUSH hbrBkGnd = CreateSolidBrush(GetSysColor(COLOR_WINDOW));
				FillRect(hdcMem, &rc, hbrBkGnd);
				DeleteObject(hbrBkGnd);
				//SetBkMode(hdcMem, TRANSPARENT);

				SelectObject(hdcMem, GetStockObject(BLACK_PEN));
				Rectangle(hdcMem, X, Y, X + W, Y + H);
				for (int x = 0; x < W - 2; x++)
				{
					int y = fan((float)x, (float)(W-3));
					SetPixel(hdcMem, X+1+x, Y+H-2-y, RGB(0, 0, 255));
				}
				for (unsigned int t = 1; t < W - 1; t++)
				{
					if (t <= history.size())
					{
						int v = history[history.size() - t];
						SetPixel(hdcMem, X + W - t - 1, Y + H - v - 2, RGB(255, 0, 0));
					}
				}

				BitBlt(ps.hdc,
					X, Y,
					W, H,
					hdcMem,
					X, Y,
					SRCCOPY);

				SelectObject(hdcMem, hbmOld);
				DeleteObject(hbmMem);
				DeleteDC(hdcMem);

				EndPaint(hWnd, &ps);
			}
			break;
		
		// exit application
		case WM_DESTROY:
			FILE * f;
			if (fopen_s(&f, pathConfig.c_str(), "wb") == 0)
			{
				fwrite(&config, sizeof(config), 1, f);
				fclose(f);
			}
			Shell_NotifyIcon(NIM_DELETE, &tray);
			PostQuitMessage(0);
			GdiplusShutdown(gdiplusToken);
			exit(0);
			break;

		default:
			return DefWindowProc(hWnd, message, wParam, lParam);
	}
	return 0;
}

int main()
{
	return wWinMain(GetModuleHandle(NULL), NULL, NULL, NULL);
}

int APIENTRY wWinMain(_In_ HINSTANCE hInstance, _In_opt_ HINSTANCE hPrevInstance, _In_ LPWSTR lpCmdLine, _In_ int nCmdShow)
{
	// enable native style
	INITCOMMONCONTROLSEX icex;
	icex.dwSize = sizeof(icex);
	icex.dwICC = ICC_BAR_CLASSES | ICC_STANDARD_CLASSES;
	InitCommonControlsEx(&icex);
	
	// load native font
	NONCLIENTMETRICS ncm;
	ncm.cbSize = sizeof(ncm);
	SystemParametersInfo(SPI_GETNONCLIENTMETRICS, sizeof(ncm), &ncm, 0);
	HFONT font = CreateFontIndirect(&ncm.lfMessageFont);

	// create interface
	LPCSTR windowClass = "WindSimulator";

	WNDCLASS wc;
	wc.style = 0; //CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc = WndProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = NULL;
	wc.hbrBackground = (HBRUSH)(COLOR_BTNFACE + 1);
	wc.hIcon = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_ICON1));
	wc.hCursor = NULL;
	wc.lpszMenuName = NULL;
	wc.lpszClassName = windowClass;
	RegisterClass(&wc);

	HWND window = CreateWindow(
		windowClass,
		"Wind Simulator",
		WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION | WS_MINIMIZEBOX,
		CW_USEDEFAULT, 0,
		500, 500,
		NULL, NULL, NULL, NULL
	);

	char labels[5][16] = {
		"Port:",
		"Speed OFF:",
		"Speed MIN:",
		"Speed MAX:",
		"Max car speed:"
	};
	int y = 20;
	for (auto &label : labels)
	{
		HWND temp = CreateWindow(
			"static",
			label,
			WS_CHILD | WS_VISIBLE | WS_TABSTOP | ES_RIGHT,
			10, y,
			100, 20,
			window, NULL, NULL, NULL
		);
		SendMessage(temp, WM_SETFONT, (WPARAM)font, TRUE);
		y += 30;
	}

	HWND combo = CreateWindow(
		WC_COMBOBOX,
		NULL,
		CBS_DROPDOWNLIST | CBS_HASSTRINGS | WS_CHILD | WS_OVERLAPPED | WS_VISIBLE,
		130, 15,
		100, 30,
		window, NULL, NULL, NULL
	);
	SendMessage(combo, WM_SETFONT, (WPARAM)font, TRUE);
	for (int i = 1; i <= 9; i++)
	{
		char A[16];
		wsprintf(A, "COM%d", i);
		SendMessage(combo, CB_ADDSTRING, 0, (LPARAM)A);
	}
	
	HWND trackbars[4];
	for (int i = 0; i < 4; i++)
	{
		auto &temp = trackbars[i];
		temp = CreateWindow(
			TRACKBAR_CLASS,
			NULL,
			WS_CHILD | WS_VISIBLE | TBS_ENABLESELRANGE | TBS_TOOLTIPS,
			120, 45 + i * 30,
			350, 30,
			window, NULL, NULL, NULL
		);
		switch (i)
		{
			case 0:
				SendMessage(temp, TBM_SETRANGE, TRUE, MAKELONG(0, 79));
				break;
			case 1:
				SendMessage(temp, TBM_SETRANGE, TRUE, MAKELONG(-20, 79));
				SendMessage(temp, TBM_SETSEL, TRUE, MAKELONG(0, 79));
				break;
			case 2:
				SendMessage(temp, TBM_SETRANGE, TRUE, MAKELONG(0, 99));
				SendMessage(temp, TBM_SETSEL, TRUE, MAKELONG(0, 79));
				break;
			case 3:
				SendMessage(temp, TBM_SETRANGE, TRUE, MAKELONG(0, 400));
				break;
		}
	}

	console = CreateWindow(
		"EDIT",
		NULL,
		WS_CHILD | WS_VISIBLE | ES_LEFT | ES_WANTRETURN | ES_AUTOVSCROLL | ES_AUTOHSCROLL | ES_MULTILINE | WS_BORDER | WS_VSCROLL | ES_READONLY,
		10, 262,
		465, 189,
		window, NULL, NULL, NULL
	);
	SendMessage(console, WM_SETFONT, (WPARAM)font, TRUE);

	// initialize GDI+
	GdiplusStartupInput gdiplusStartupInput;
	GdiplusStartup(&gdiplusToken, &gdiplusStartupInput, NULL);

	// tray icon
	tray.cbSize = sizeof(tray);
	tray.hIcon = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_ICON1));
	tray.hWnd = window;
	strcpy_s(tray.szTip, "Wind Simulator");
	tray.uCallbackMessage = WM_APP + 1;
	tray.uFlags = NIF_ICON | NIF_TIP | NIF_MESSAGE;
	tray.uID = 0;
	Shell_NotifyIcon(NIM_ADD, &tray);

	// show the window?
	if (GetConsoleWindow())
	{
		ShowWindow(window, SW_RESTORE);
		UpdateWindow(window);
	}

	// initialize shared memories
	for (auto &share : shares)
	{
		auto handle = CreateFileMapping(INVALID_HANDLE_VALUE, NULL, PAGE_READWRITE, 0, OVERALLOCATE, share.name);
		if (!handle)
		{
			error("CreateFileMapping failed: " + string(share.name));
		}
		share.pointer = (unsigned char*)MapViewOfFile(handle, FILE_MAP_READ, 0, 0, share.size);
		if (!share.pointer)
		{
			error("MapViewOfFile failed: " + string(share.name));
		}
	}

	// map shared memories
	SPageFilePhysics *ac_physics = (SPageFilePhysics*)shares[AC_PHYSICS].pointer;
	SPageFileGraphic *ac_graphic = (SPageFileGraphic*)shares[AC_GRAPHICS].pointer;
	SPageFileStatic *ac_static = (SPageFileStatic*)shares[AC_STATIC].pointer;
	SharedMemory *pcars2 = (SharedMemory*)shares[PCARS2].pointer;
	r3e_shared *r3e = (r3e_shared*)shares[R3E].pointer;

	// read config
	TCHAR szPath[MAX_PATH];
	if (!SUCCEEDED(SHGetFolderPath(NULL, CSIDL_APPDATA | CSIDL_FLAG_CREATE, NULL, 0, szPath)))
	{
		error("SHGetFolderPath failed: %APPDATA%");
	}
	string path = string(szPath) + "/WindSimulator";
	CreateDirectory(path.c_str(), NULL);
	pathConfig = path + "/config.bin";
	FILE *f;
	if (fopen_s(&f, pathConfig.c_str(), "rb") == 0)
	{
		fread(&config, sizeof(config), 1, f);
		fclose(f);
	}
	SendMessage(combo, CB_SETCURSEL, config.com_port, 0);
	SendMessage(trackbars[0], TBM_SETPOS, TRUE, config.fan_off);
	SendMessage(trackbars[1], TBM_SETPOS, TRUE, config.fan_min);
	SendMessage(trackbars[2], TBM_SETPOS, TRUE, config.fan_max);
	SendMessage(trackbars[3], TBM_SETPOS, TRUE, config.car_spd);

	// car parameters
	string previousCombo = "";
	float maxSpeed;
	bool newMaxSpeed = false;
	int automatic = -1;

	// polling
	while (true)
	{
		// handle messages
		MSG msg;
		while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
		
		// reset
		string currentCombo = "";
		float currentSpeed = 0;

		// is Assetto Corsa running?
		if (ac_graphic->status == AC_LIVE)
		{
			char carModel[99], track[99], trackConfiguration[99];
			WideCharToMultiByte(CP_UTF8, 0, ac_static->carModel, -1, carModel, 99, NULL, NULL);
			WideCharToMultiByte(CP_UTF8, 0, ac_static->track, -1, track, 99, NULL, NULL);
			WideCharToMultiByte(CP_UTF8, 0, ac_static->trackConfiguration, -1, trackConfiguration, 99, NULL, NULL);

			currentCombo = string(carModel) + " @ " + string(track) + " " + string(trackConfiguration);
			currentSpeed = ac_physics->speedKmh;
		}

		// is Project Cars 2 running?
		if (pcars2->mGameState == GAME_INGAME_PLAYING)
		{
			currentCombo = string(pcars2->mCarName) + " @ " + string(pcars2->mTrackLocation) + " " + string(pcars2->mTrackVariation);
			currentSpeed = pcars2->mSpeed * 3.6f;
		}

		// is RaceRoom Racing Experience running?
		if (r3e->session_phase == R3E_SESSION_PHASE_GREEN)
		{
			currentCombo = string((char*) r3e->vehicle_info.name) + " @ " + string((char*) r3e->track_name) + " " + string((char*) r3e->layout_name);
			currentSpeed = r3e->car_speed * 3.6f;
		}

		// update settings
		config.fan_off = (int) SendMessage(trackbars[0], TBM_GETPOS, 0, 0);
		config.fan_min = (int) SendMessage(trackbars[1], TBM_GETPOS, 0, 0);
		config.fan_max = (int) SendMessage(trackbars[2], TBM_GETPOS, 0, 0);
		config.car_spd = (int) SendMessage(trackbars[3], TBM_GETPOS, 0, 0);

		// update mode
		int test = config.car_spd == 0;
		if (test != automatic)
		{
			automatic = test;
			log(automatic ? "Automatic maximum car speed detection" : "Manual maximum car speed");
		}

		// default speed: OFF
		char fanSpeed = config.fan_off;

		// is a game running?
		if (currentCombo != "")
		{
			// car changed?
			if (currentCombo != previousCombo)
			{
				log("New combo: " + currentCombo);
				previousCombo = currentCombo;
				maxSpeed = 0;
			}

			// new max speed?
			if (currentSpeed > maxSpeed)
			{
				maxSpeed = currentSpeed;
				newMaxSpeed = true;
			}

			// did we reach max speed during previous iteration?
			else
			{
				if (newMaxSpeed && maxSpeed > 1 && automatic)
				{
					log("New max speed: " + to_string(maxSpeed));
				}
				newMaxSpeed = false;
			}

			// set fan speed
			fanSpeed = fan(currentSpeed, automatic ? maxSpeed : config.car_spd);
		}

		// save history
		history.push_back(fanSpeed);
		if (history.size() > W)
		{
			history.pop_front();
		}

		// update chart
		RECT rect = { X, Y, X+W, Y+H };
		InvalidateRect(window, &rect, FALSE);
		
		// send data to Arduino
		if (SP == NULL || !SP->IsConnected() || !SP->WriteData(&fanSpeed, 1))
		{
			// try to reconnect
			char comString[20];
			sprintf_s(comString, "\\\\.\\COM%d", config.com_port + 1);

			// were we connecte before?
			if (SP != NULL && SP->IsConnected())
			{
				log("Disconnected from COM" + to_string(config.com_port + 1));
			}
			delete SP;
			SP = new Serial(comString);
			if (SP->IsConnected())
			{
				log("Connected to COM" + to_string(config.com_port + 1));
			}
		}

		// 10Hz
		this_thread::sleep_for(chrono::milliseconds(100));
	}

	return 0;
}
